import cv2

from gbvision.utils.pipeline import PipeLine


class FeedWindow:
    """
    a basic window class
    """
    def __init__(self, drawing_pipeline=PipeLine(), exit_key: str='qQ', window_name='feed'):
        """
        initializes the window
        :param drawing_pipeline: optional, a pre-processing pipeline that draws on the frame before displaying it
        :param exit_key: an array of chars (a string), which when pressed the window will stop displaying
        :param window_name: the title of the window
        """
        self.drawing_pipeline = drawing_pipeline
        self.exit_key = exit_key
        self.window_name = window_name

    def open(self, flags=cv2.WINDOW_FREERATIO):
        """
        initializes the window
        :param flags: some opencv window flags
        """
        cv2.namedWindow(self.window_name, flags)

    def show_frame(self, frame) -> bool:
        """
        shows the frame
        :param frame: the frame to show
        :return: True if the window is still open, False otherwise
        """
        cv2.imshow(self.window_name, self.drawing_pipeline(frame))
        k = chr(cv2.waitKey(1) & 0xFF)
        if k in self.exit_key:
            self.close()
            return False
        return True

    def close(self):
        """
        closes this window
        """
        cv2.destroyWindow(self.window_name)





def crop(frame, x, y, w, h):
    """
    crops the image from (x, y) to (x+w, y+h)
    :param frame: the frame to crop
    :param x: the x coordinate to crop from
    :param y: the y coordinate to crop from
    :param w: the width of the cropped image
    :param h: the height of the cropped image
    :return: the cropped image
    """
    return frame[y:y + h, x:x + w, :]



from gbvision.utils.pipeline import PipeLine


def circle_collision(center1, r1, center2, r2):
    return (center1[0] - center2[0]) ** 2 + (center1[1] - center2[1]) ** 2 < (r1 + r2) ** 2


def _filter_inner_circles(circles):
    filtered_circles = []
    for i, circle in enumerate(circles):
        circle_invalid = False
        for j in range(i):
            circle_invalid = circle_collision(circle[0], circle[1], circles[j][0], circles[j][1])
            if circle_invalid:
                break
        if not circle_invalid:
            filtered_circles.append(circle)

    return filtered_circles


filter_inner_circles = PipeLine(_filter_inner_circles)


def rect_collision(r1, r2):
    return not (r1[0] > r2[0] + r2[2] or
                r1[0] + r1[2] < r1[0] or
                r1[1] > r2[1] + r2[3] or
                r1[1] + r1[3] > r2[1])


def _filter_inner_rects(rects):
    filtered_rects = []
    for i, rect in enumerate(rects):
        rect_invalid = False
        for j in range(i):
            rect_invalid = rect_collision(rect, rects[j])
            if rect_invalid:
                break
        if not rect_invalid:
            filtered_rects.append(rect)
    return filtered_rects


filter_inner_rects = PipeLine(_filter_inner_rects)



from gbvision.exceptions import AbstractMethodCallingException


class StreamBroadcaster:
    """
    this is an abstract broadcaster that sends stream to a broadcast receiver
    this class should not be instanced but inherited from
    """

    def __init__(self, fx: float = 1.0, fy: float = 1.0, use_grayscale: bool = False, max_fps: int = None):
        """
        creates a new stream broadcaster with all parameters that are used in every broadcaster
        :param fx: ratio between width of the given frame to the width of the frame sent, default is 1 (same width)
        :param fy: ratio between height of the given frame to the height of the frame sent, default is 1 (same height)
        :param use_grayscale: boolean indicating if the frame should be converted to grayscale when sent,
         default is False
        :param max_fps: integer representing the maximum fps (frames per second) of the stream, when set to None
        there is no fps limitation, default is None
        """
        self.fx = fx
        self.fy = fy
        self.use_grayscale = use_grayscale
        self.max_fps = max_fps

    def send_frame(self, frame):
        """
        sends the given frame to the stream receiver
        :param frame: the frame to send
        """
        raise AbstractMethodCallingException()



import functools


class PipeLine:
    """
    a class representing a pipeline of function
    each function receives one input, which is the output of the previous function in the pipeline
    pipelines are great for representing a long computer vision function (which is why such functions
    are called pipelines)
    """
    def __init__(self, *functions):
        """
        initializes this pipeline
        :param functions: a tuple of functions to run one after the other as a pipeline
        """
        self.functions = list(functions)

    def __call__(self, image):
        """
        activate this pipeline and return the result
        :param image: the input to the first function in the pipeline (also the input to the entire pipeline)
        doesn't have to be an image, can be anything
        :return: the output of the last function in the pipeline, can be data type
        """
        return functools.reduce(lambda x, f: f(x), self.functions, image)

    def __add__(self, other):
        """
        creates a new pipeline which uses the output of this pipeline as input to the other pipeline
        :param other: the second pipeline
        :return: a new pipeline, and calling this pipeline with the parameter image is similar to
        performing other(self(image))
        """
        if isinstance(other, PipeLine):
            return PipeLine(*self.functions + other.functions)
        return PipeLine(*self.functions + [other])

    def __iadd__(self, fun):
        """
        adds a function or pipeline to the end of this pipeline (changes this pipeline, doesn't return a new one)
        :param fun: the function or pipeline
        """
        if isinstance(fun, PipeLine):
            self.functions += fun.functions
        self.functions.append(fun)

    def __getitem__(self, item):
        """
        gets the function at index item
        :param item: the index
        :return: the item
        """
        return self.functions[item]

    def __setitem__(self, key, value):
        """
        sets the function at index key to the new function value
        :param key: the index
        :param value: the new function
        """
        self.functions[key] = value

    def __iter__(self):
        """
        :return: an iterator that iterates through all functions in this pipeline
        """
        return iter(self.functions)



# constants
from .constants.net import *
from .constants.cameras import LIFECAM_3000, LIFECAM_STUDIO, GENERIC_CAMERA
from .constants.math import EPSILON, SQRT_PI
from .constants.system import CONTOURS_INDEX, cv_config

# exceptions
from .exceptions.vision_warning import VisionWarning
from .exceptions.vision_exception import VisionException
from .exceptions.could_not_read_frame_exception import CouldNotReadFrameException
from .exceptions.device_not_found_exception import DeviceNotFoundException
from .exceptions.abstract_method_calling_exception import AbstractMethodCallingException

# finders
from .finders.object_finder import ObjectFinder
from .finders.circle_finder import CircleFinder
from .finders.rect_finder import RectFinder
from .finders.polygon_finder import PolygonFinder
from .finders.rotated_rect_finder import RotatedRectFinder
from .finders.target_pair_finder import TargetPairFinder

# gui
from .gui.window import Window
from .gui.stream_window import StreamWindow
from .gui.feed_window import FeedWindow
from .gui.camera_window import CameraWindow

# models
from .models.basic_ops import corners, edges, sharpen, blur, blue, red, green, gray
from .models.contours import contour_center, contours_centers, contours_to_circles, contours_to_circles_sorted, \
    contours_to_ellipses, contours_to_ellipses_sorted, contours_to_polygons, contours_to_rects, \
    contours_to_rects_sorted, contours_to_rotated_rects, contours_to_rotated_rects_sorted, filter_contours, \
    find_contours, sort_contours
from .models.shapes import circle_collision, filter_inner_circles, filter_inner_rects, rect_collision

# net
from .net.stream_receiver import StreamReceiver
from .net.stream_broadcaster import StreamBroadcaster
from .net.tcp_stream_broadcaster import TCPStreamBroadcaster
from .net.tcp_stream_receiver import TCPStreamReceiver
from .net.udp_stream_broadcaster import UDPStreamBroadcaster
from .net.udp_stream_receiver import UDPStreamReceiver

# tools
from .tools.list_tools import split_list
from .tools.image_tools import crop
from .tools.finding_tools import distance_from_object, angle_by_location2d, angle_by_location3d

# utils
from .utils.camera import Camera
from .utils.stream_camera import StreamCamera
from .utils.camera_data import CameraData
from .utils.usb_camera import USBCamera
from .utils.usb_stream_camera import USBStreamCamera
from .utils.async_usb_camera import AsyncUSBCamera
from .utils.threshold import Threshold
from .utils.threshold_group import ThresholdGroup
from .utils.pipeline import PipeLine
from .utils.game_object import GameObject


cv_config()



import cv2


def hls_threshold(frame, params):
    """
    thresholds the image according to HLS values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [hmin hmax]
                                                 [lmin lmax]
                                                 [smin smax]
    :return: binary threshold image
    """

    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2HLS)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def hsv_threshold(frame, params):
    """
    thresholds the image according to HSV values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [hmin hmax]
                                                 [lmin lmax]
                                                 [smin smax]
    :return: binary threshold image
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def rgb_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [hmin hmax]
                                                 [lmin lmax]
                                                 [smin smax]
    :return: binary threshold image
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def bgr_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [hmin hmax]
                                                 [lmin lmax]
                                                 [smin smax]
    :return: binary threshold image
    """
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def luv_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [hmin hmax]
                                                 [lmin lmax]
                                                 [smin smax]
    :return: binary threshold image
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2LUV)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def lab_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [hmin hmax]
                                                 [lmin lmax]
                                                 [smin smax]
    :return: binary threshold image
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2LAB)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def yuv_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [hmin hmax]
                                                 [lmin lmax]
                                                 [smin smax]
    :return: binary threshold image
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2YUV)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def xyz_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [hmin hmax]
                                                 [lmin lmax]
                                                 [smin smax]
    :return: binary threshold image
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2XYZ)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


THRESHOLD_NAME_TABLE = {
    'BGR': bgr_threshold,
    'RGB': rgb_threshold,
    'HLS': hls_threshold,
    'HSV': hsv_threshold,
    'LUV': luv_threshold,
    'LAB': lab_threshold,
    'YUV': yuv_threshold,
    'XYZ': xyz_threshold
}


class Threshold:
    def __init__(self, lst, thresh_type='HSV'):
        assert thresh_type.upper() in THRESHOLD_NAME_TABLE
        self.init = lst
        self.type = thresh_type.upper()

    def __len__(self):
        return len(self.init)

    def __getitem__(self, item):
        return self.init[item]

    def __setitem__(self, key, value):
        self.init[key] = value

    def __iter__(self):
        return iter(self.init)

    def __call__(self, frame):
        return THRESHOLD_NAME_TABLE[self.type](frame, self.init)




from gbvision.exceptions import AbstractMethodCallingException
from gbvision.utils.game_object import GameObject, Camera


class ObjectFinder:
    """
    this is an abstract class that represents an object finder
    an object finder is a type that outputs an object's 3d real location based on an of it image it's
    GameObject real-life parameters
    """
    def __init__(self, threshold_func, game_object: GameObject):
        """
        initializes the finder
        :param threshold_func: a pipeline (or any sort of function) that returns a binary threshold of the object
        the finder is searching, the object needs to be white and the rest if the image black (doesn't
        have to be perfect)
        :param game_object: the game object descriptor for the real-life parameters of the finder's target
        """
        self.threshold = threshold_func
        self.game_object = game_object

    def __call__(self, frame, camera: Camera):
        """
        finds all instances of the object in the frame
        :param frame:
        :param camera:
        :return:
        """
        raise AbstractMethodCallingException()



from gbvision.net.stream_broadcaster import StreamBroadcaster
from .camera import Camera, AbstractMethodCallingException


class StreamCamera(Camera):
    """
    an abstract class that represents a streaming camera
    the streaming camera is very similar to a regular camera, but has an option
    which allows it to stream the frames when it reads them
    """

    def is_streaming(self) -> bool:
        """
        checks if the camera is currently streaming
        :return: True if camera is streaming, otherwise Fasle
        """
        raise AbstractMethodCallingException()

    def toggle_stream(self, should_stream: bool):
        """
        turn on or off the stream feature
        :param should_stream: True to activate stream, False to deactivate
        """
        raise AbstractMethodCallingException()

    @staticmethod
    def create_type(camera_class):
        """
        creates a new class that is similar to the given class, but has the stream feature
        the constructor of the new class is the same as the given class, but adds a new parameter at the beginning
        the first parameter to the new constructor is a stream broadcaster
        :param camera_class: the class to wrap
        :return: the wrapped class
        """

        class _StreamCamera(camera_class, StreamCamera):
            def __init__(self, broadcaster: StreamBroadcaster, *args, **kwargs):
                camera_class.__init__(self, *args, **kwargs)
                self._is_streaming = False
                self.stream_broadcaster = broadcaster

            def is_streaming(self):
                return self._is_streaming

            def toggle_stream(self, should_stream: bool):
                self._is_streaming = should_stream

            def read(self, image=None):
                ok, frame = camera_class.read(self, image=image)
                if self._is_streaming:
                    self.stream_broadcaster.send_frame(frame)
                return ok, frame

        return _StreamCamera



from functools import reduce

import sys
import cv2


class ThresholdGroup:
    def __init__(self, *thresholds, **kwargs):
        self.binary_mask = cv2.bitwise_or
        self.default_pixel = 0
        if 'binary_mask' in kwargs:
            self.binary_mask = kwargs['binary_mask']
            del kwargs['binary_mask']
        if 'default_pixel' in kwargs:
            self.default_pixel = kwargs['default_pixel']
            del kwargs['default_pixel']
        for i in kwargs:
            print('[WARN] keyword value %s is never used' % i, file=sys.stderr)
        self.thresholds = list(thresholds)

    def __call__(self, frame):
        return reduce(lambda th_frame, threshold: self.binary_mask(th_frame, threshold(frame)), self.thresholds, 0)

    def __add__(self, other):
        if isinstance(other, ThresholdGroup):
            return ThresholdGroup(self.thresholds + other.thresholds)
        return ThresholdGroup(self.thresholds + [other])

    def __iter__(self):
        return iter(self.thresholds)

    def __len__(self):
        return len(self.thresholds)

    def __getitem__(self, item):
        return self.thresholds[item]

    def __setitem__(self, key, value):
        self.thresholds[key] = value

    def __iadd__(self, other):
        if isinstance(other, ThresholdGroup):
            self.thresholds += other.thresholds
        else:
            self.thresholds += [other]
        return self



from .basic_ops import corners, edges, sharpen, blur, blue, red, green, gray
from .contours import contour_center, contours_centers, contours_to_circles, contours_to_circles_sorted, \
    contours_to_ellipses, contours_to_ellipses_sorted, contours_to_polygons, contours_to_rects, \
    contours_to_rects_sorted, contours_to_rotated_rects, contours_to_rotated_rects_sorted, filter_contours, \
    find_contours, sort_contours
from .shapes import circle_collision, filter_inner_circles, filter_inner_rects, rect_collision



from threading import Thread

from gbvision.exceptions import AbstractMethodCallingException
from gbvision.utils.pipeline import PipeLine


class Window:
    """
    an abstract window class
    """
    def __init__(self, window_name: str, exit_button='qQ', drawing_pipeline=PipeLine()):
        """
        initializes the window
        :param window_name: the title of the new window
        :param exit_button: the button that closes the window
        :param drawing_pipeline: optional, a pipeline that draws on each frame before displaying it
        """
        self.window_name = window_name
        self.exit_button = exit_button
        self.drawing_pipeline = drawing_pipeline

    def show(self):
        raise AbstractMethodCallingException()

    def show_async(self):
        Thread(target=self.show).start()



import numpy as np


def distance_from_object(loc) -> float:
    """
    the absolute distance from the camera to this object
    :param loc: the object's location (2d or 3d)
    :returns: the absolute distance (float)
    """
    if len(loc) <= 3:
        return np.linalg.norm(loc)
    return np.linalg.norm(loc[:3])


def angle_by_location2d(loc) -> float:
    """
    calculates the angle from the camera to the object
    :param loc: the 2d location
    :return: the angle (in radians)
    """
    return np.arctan(loc[0] / loc[1])


def angle_by_location3d(loc) -> float:
    """
    calculates the angle from the camera to the object
    :param loc: the 3d location
    :return: the angle (in radians)
    """
    return np.arctan(loc[0] / loc[2])
